#include <vm_lifters.hpp>

namespace vm
{
    llvm::Value *lifters_t::and_flags( vm::devirt_t *rtn, std::uint8_t byte_size, llvm::Value *result )
    {
        auto cf = llvm::ConstantInt::get( llvm::IntegerType::get( *rtn->llvm_ctx, 64 ), 0 );
        auto of = llvm::ConstantInt::get( llvm::IntegerType::get( *rtn->llvm_ctx, 64 ), 0 );

        auto sf = rtn->compute_sf( byte_size, result );
        auto zf = rtn->compute_zf( byte_size, result );
        auto pf = llvm::ConstantInt::get( llvm::IntegerType::get( *rtn->llvm_ctx, 64 ), 0 );

        return rtn->combine_flags( cf, pf, llvm::ConstantInt::get( llvm::IntegerType::get( *rtn->llvm_ctx, 64 ), 0 ),
                                   zf, sf, of );
    }

    lifters_t::lifter_callback_t lifters_t::nandq =
        [ & ]( vm::devirt_t *rtn, const vm::instrs::code_block_t &vm_code_block,
               const vm::instrs::virt_instr_t &vinstr, llvm::IRBuilder<> *ir_builder ) {
            auto t1 = rtn->pop( 8 );
            auto t2 = rtn->pop( 8 );

            auto t1_not = ir_builder->CreateNot( t1 );
            auto t2_not = ir_builder->CreateNot( t2 );

            auto t3 = ir_builder->CreateAnd( { t1_not, t2_not } );
            rtn->push( 8, t3 );

            auto &vmp_rtn = rtn->vmp_rtns.back();
            auto flags = and_flags( rtn, 8, t3 );
            ir_builder->CreateStore( flags, vmp_rtn->flags );
            rtn->push( 8, rtn->load_value( 8, vmp_rtn->flags ) );
        };

    lifters_t::lifter_callback_t lifters_t::nanddw =
        [ & ]( vm::devirt_t *rtn, const vm::instrs::code_block_t &vm_code_block,
               const vm::instrs::virt_instr_t &vinstr, llvm::IRBuilder<> *ir_builder ) {
            auto t1 = rtn->pop( 4 );
            auto t2 = rtn->pop( 4 );

            auto t1_not = ir_builder->CreateNot( t1 );
            auto t2_not = ir_builder->CreateNot( t2 );

            auto t3 = ir_builder->CreateAnd( { t1_not, t2_not } );
            rtn->push( 4, t3 );

            auto &vmp_rtn = rtn->vmp_rtns.back();
            auto flags = and_flags( rtn, 4, t3 );
            ir_builder->CreateStore( flags, vmp_rtn->flags );
            rtn->push( 8, rtn->load_value( 8, vmp_rtn->flags ) );
        };

} // namespace vm